import paramiko
import os
from os.path import dirname
import stat


def remote_isdir(attr):
    return stat.S_ISDIR(attr.st_mode)


def mkdir_p(sftp, remote_directory):
    print("Trying to create %s folder" % remote_directory)
    """Change to this directory, recursively making new folders if needed.
    Returns True if any folders were created."""
    if remote_directory == '/':
        # absolute path so change directory to root
        sftp.chdir('/')
        return
    if remote_directory == '':
        # top-level relative directory must exist
        return
    try:
        sftp.chdir(remote_directory) # sub-directory exists
    except IOError:
        dirname, basename = os.path.split(remote_directory.rstrip('/'))
        mkdir_p(sftp, dirname) # make parent directories
        sftp.mkdir(basename) # sub-directory missing, so created it
        sftp.chdir(basename)
        return True


class SSH:
    def __init__(self, user, host, remote_scratch=""):
        # Init
        self.user = user
        self.host = host
        self._host_config_id = None
        self._ssh = None
        self._ssh_config = None
        self._user_config_file = None
        self._host_config = None
        self.sftp = None
        self.remote_scratch = remote_scratch

        # SSH
        self.connect()
        self.init_transport()

    def init_transport(self):
        self.sftp = self._ssh.open_sftp()

    def get(self, remote_path, local_path):
        if not self.sftp:
            self.init_transport()
        self.sftp.get(remote_path, local_path)

    def put(self, local_path, remote_path):
        if not self.sftp:
            self.init_transport()
        remote_folder_path = dirname(remote_path)
        try:
            self.sftp.chdir(remote_folder_path)  # Test if remote_path exists
        except IOError:
            mkdir_p(self.sftp, remote_folder_path)
        self.sftp.put(local_path, remote_path)

    def write_file(self, text, remote_path):
        if not self.sftp:
            self.init_transport()
        remote_folder_path = dirname(remote_path)
        try:
            self.sftp.chdir(remote_folder_path)  # Test if remote_path exists
        except IOError:
            mkdir_p(self.sftp, remote_folder_path)
        remote_file = self.sftp.open(remote_path, "w")
        remote_file.write(text)

    def connect(self):
            """
            Creates ssh connection to host

            :return: True if connection is created, False otherwise
            :rtype: bool
            """
            try:
                self._ssh = paramiko.SSHClient()
                self._ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                self._ssh_config = paramiko.SSHConfig()
                self._user_config_file = os.path.expanduser("~/.ssh/config")
                if os.path.exists(self._user_config_file):
                    with open(self._user_config_file) as f:
                        # noinspection PyTypeChecker
                        self._ssh_config.parse(f)
                self._host_config = self._ssh_config.lookup(self.host)
                if 'identityfile' in self._host_config:
                    self._host_config_id = self._host_config['identityfile']
                if 'proxycommand' in self._host_config:
                    self._proxy = paramiko.ProxyCommand(self._host_config['proxycommand'])
                    self._ssh.connect(self._host_config['hostname'], 22, username=self.user,
                                      key_filename=self._host_config_id, sock=self._proxy)
                else:
                    self._ssh.connect(self._host_config['hostname'], 22, username=self.user,
                                      key_filename=self._host_config_id)
                return True
            except IOError as e:
                print('Can not create ssh connection to {0}: {1}', self.host, e.strerror)
                raise e

    def execute(self, command):
        return self._ssh.exec_command(command)

    def list_dir(self, dir_path):
        return self.sftp.listdir(dir_path)

    def is_remote_file(self, path):
        """Check if the path on the remote server is a file."""
        try:
            file_attr = self.sftp.stat(path)
            # Check if the path is a regular file
            return stat.S_ISREG(file_attr.st_mode)
        except IOError as e:
            # If an IOError occurs, the file does not exist or there's an error accessing it
            return False

